@@ -20,7 +20,6 @@
 static void install_library_entry PROTO((ptr n, ptr x));
 static void scheme_install_library_entry PROTO((void));
 static void create_library_entry_vector PROTO((void));
-static void install_c_entry PROTO((iptr i, ptr x));
 static void create_c_entry_vector PROTO((void));
 static void s_instantiate_code_object PROTO((void));
 static void s_link_code_object PROTO((ptr co, ptr objs));
@@ -46,6 +45,10 @@ static void install_library_entry(n, x) ptr n, x; {
     if (n == FIX(library_cpu_features))
       x86_64_set_popcount_present(x);
 #endif
+#ifdef PORTABLE_BYTECODE_BIGENDIAN
+    if (n == FIX(library_dounderflow))
+      S_swap_dounderflow_header_endian(CLOSCODE(x));
+#endif
 }
 
 ptr S_lookup_library_entry(n, errorp) iptr n; IBOOL errorp; {
@@ -82,10 +85,10 @@ ptr int2ptr(iptr f)
    return (ptr)(f & ~0x3);
 }
 #else /* HPUX */
-#define proc2ptr(x) (ptr)(iptr)(x)
+#define proc2ptr(x) TO_PTR(x)
 #endif /* HPUX */
 
-static void install_c_entry(i, x) iptr i; ptr x; {
+void S_install_c_entry(i, x) iptr i; ptr x; {
     if (i < 0 || i >= c_entry_vector_size)
         S_error1("install_c_entry", "invalid index ~s", FIX(i));
     if (Svector_ref(S_G.c_entry_vector, i) != Sfalse)
@@ -116,41 +119,50 @@ static void create_c_entry_vector() {
     for (i = 0; i < c_entry_vector_size; i++)
         INITVECTIT(S_G.c_entry_vector, i) = Sfalse;
 
-    install_c_entry(CENTRY_thread_context, proc2ptr(S_G.thread_context));
-    install_c_entry(CENTRY_get_thread_context, proc2ptr(s_get_thread_context));
-    install_c_entry(CENTRY_handle_apply_overflood, proc2ptr(S_handle_apply_overflood));
-    install_c_entry(CENTRY_handle_docall_error, proc2ptr(S_handle_docall_error));
-    install_c_entry(CENTRY_handle_overflow, proc2ptr(S_handle_overflow));
-    install_c_entry(CENTRY_handle_overflood, proc2ptr(S_handle_overflood));
-    install_c_entry(CENTRY_handle_nonprocedure_symbol, proc2ptr(S_handle_nonprocedure_symbol));
-    install_c_entry(CENTRY_thread_list, (ptr)&S_threads);
-    install_c_entry(CENTRY_split_and_resize, proc2ptr(S_split_and_resize));
+    S_install_c_entry(CENTRY_thread_context, proc2ptr(S_G.thread_context));
+    S_install_c_entry(CENTRY_get_thread_context, proc2ptr(s_get_thread_context));
+    S_install_c_entry(CENTRY_handle_apply_overflood, proc2ptr(S_handle_apply_overflood));
+    S_install_c_entry(CENTRY_handle_docall_error, proc2ptr(S_handle_docall_error));
+    S_install_c_entry(CENTRY_handle_overflow, proc2ptr(S_handle_overflow));
+    S_install_c_entry(CENTRY_handle_overflood, proc2ptr(S_handle_overflood));
+    S_install_c_entry(CENTRY_handle_nonprocedure_symbol, proc2ptr(S_handle_nonprocedure_symbol));
+    S_install_c_entry(CENTRY_thread_list, TO_PTR(&S_threads));
+    S_install_c_entry(CENTRY_split_and_resize, proc2ptr(S_split_and_resize));
 #ifdef PTHREADS
-    install_c_entry(CENTRY_raw_collect_cond, (ptr)&S_collect_cond);
-    install_c_entry(CENTRY_raw_collect_thread0_cond, (ptr)&S_collect_thread0_cond);
-    install_c_entry(CENTRY_raw_tc_mutex, (ptr)&S_tc_mutex);
-    install_c_entry(CENTRY_activate_thread, proc2ptr(S_activate_thread));
-    install_c_entry(CENTRY_deactivate_thread, proc2ptr(Sdeactivate_thread));
-    install_c_entry(CENTRY_unactivate_thread, proc2ptr(S_unactivate_thread));
+    S_install_c_entry(CENTRY_raw_collect_cond, TO_PTR(&S_collect_cond));
+    S_install_c_entry(CENTRY_raw_collect_thread0_cond, TO_PTR(&S_collect_thread0_cond));
+    S_install_c_entry(CENTRY_raw_tc_mutex, TO_PTR(&S_tc_mutex));
+    S_install_c_entry(CENTRY_raw_terminated_cond, TO_PTR(&S_terminated_cond));
+    S_install_c_entry(CENTRY_activate_thread, proc2ptr(S_activate_thread));
+    S_install_c_entry(CENTRY_deactivate_thread, proc2ptr(Sdeactivate_thread));
+    S_install_c_entry(CENTRY_unactivate_thread, proc2ptr(S_unactivate_thread));
 #endif /* PTHREADS */
-    install_c_entry(CENTRY_handle_values_error, proc2ptr(S_handle_values_error));
-    install_c_entry(CENTRY_handle_mvlet_error, proc2ptr(S_handle_mvlet_error));
-    install_c_entry(CENTRY_handle_arg_error, proc2ptr(S_handle_arg_error));
-    install_c_entry(CENTRY_handle_event_detour, proc2ptr(S_handle_event_detour));
-    install_c_entry(CENTRY_foreign_entry, proc2ptr(S_foreign_entry));
-    install_c_entry(CENTRY_install_library_entry, proc2ptr(scheme_install_library_entry));
-    install_c_entry(CENTRY_get_more_room, proc2ptr(S_get_more_room));
-    install_c_entry(CENTRY_scan_remembered_set, proc2ptr(S_scan_remembered_set));
-    install_c_entry(CENTRY_instantiate_code_object, proc2ptr(s_instantiate_code_object));
-    install_c_entry(CENTRY_Sreturn, proc2ptr(S_return));
-    install_c_entry(CENTRY_Scall_one_result, proc2ptr(S_call_one_result));
-    install_c_entry(CENTRY_Scall_any_results, proc2ptr(S_call_any_results));
-    install_c_entry(CENTRY_segment_info, proc2ptr(S_segment_info));
-    install_c_entry(CENTRY_bignum_mask_test, proc2ptr(S_bignum_mask_test));
+    S_install_c_entry(CENTRY_handle_values_error, proc2ptr(S_handle_values_error));
+    S_install_c_entry(CENTRY_handle_mvlet_error, proc2ptr(S_handle_mvlet_error));
+    S_install_c_entry(CENTRY_handle_arg_error, proc2ptr(S_handle_arg_error));
+    S_install_c_entry(CENTRY_handle_event_detour, proc2ptr(S_handle_event_detour));
+    S_install_c_entry(CENTRY_foreign_entry, proc2ptr(S_foreign_entry));
+    S_install_c_entry(CENTRY_install_library_entry, proc2ptr(scheme_install_library_entry));
+    S_install_c_entry(CENTRY_get_more_room, proc2ptr(S_get_more_room));
+    S_install_c_entry(CENTRY_scan_remembered_set, proc2ptr(S_scan_remembered_set));
+    S_install_c_entry(CENTRY_instantiate_code_object, proc2ptr(s_instantiate_code_object));
+    S_install_c_entry(CENTRY_Sreturn, proc2ptr(S_return));
+    S_install_c_entry(CENTRY_Scall_one_result, proc2ptr(S_call_one_result));
+    S_install_c_entry(CENTRY_Scall_any_results, proc2ptr(S_call_any_results));
+    S_install_c_entry(CENTRY_segment_info, proc2ptr(S_segment_info));
+    S_install_c_entry(CENTRY_bignum_mask_test, proc2ptr(S_bignum_mask_test));
+    S_install_c_entry(CENTRY_null_immutable_vector, TO_PTR(S_G.null_immutable_vector));
+    S_install_c_entry(CENTRY_null_immutable_bytevector, TO_PTR(S_G.null_immutable_bytevector));
+    S_install_c_entry(CENTRY_null_immutable_string, TO_PTR(S_G.null_immutable_string));
+}
+
+void S_check_c_entry_vector() {
+    INT i;
 
     for (i = 0; i < c_entry_vector_size; i++) {
 #ifndef PTHREADS
-      if (i == CENTRY_raw_collect_cond || i == CENTRY_raw_tc_mutex
+      if (i == CENTRY_raw_collect_cond || i == CENTRY_raw_collect_thread0_cond
+          || i == CENTRY_raw_tc_mutex || i == CENTRY_raw_terminated_cond
           || i == CENTRY_activate_thread || i == CENTRY_deactivate_thread
           || i == CENTRY_unactivate_thread)
         continue;
@@ -170,6 +182,7 @@ void S_prim_init() {
 
     Sforeign_symbol("(cs)fixedpathp", (void *)S_fixedpathp);
     Sforeign_symbol("(cs)bytes_allocated", (void *)S_compute_bytes_allocated);
+    Sforeign_symbol("(cs)bytes_finalized", (void *)S_bytes_finalized);
     Sforeign_symbol("(cs)curmembytes", (void *)S_curmembytes);
     Sforeign_symbol("(cs)maxmembytes", (void *)S_maxmembytes);
     Sforeign_symbol("(cs)resetmaxmembytes", (void *)S_resetmaxmembytes);
@@ -188,6 +201,8 @@ void S_prim_init() {
     Sforeign_symbol("(cs)maxgen", (void *)S_maxgen);
     Sforeign_symbol("(cs)set_maxgen", (void *)S_set_maxgen);
     Sforeign_symbol("(cs)minfreegen", (void *)S_minfreegen);
+    Sforeign_symbol("(cs)set_minmarkgen", (void *)S_set_minmarkgen);
+    Sforeign_symbol("(cs)minmarkgen", (void *)S_minmarkgen);
     Sforeign_symbol("(cs)set_minfreegen", (void *)S_set_minfreegen);
     Sforeign_symbol("(cs)enable_object_counts", (void *)S_enable_object_counts);
     Sforeign_symbol("(cs)set_enable_object_counts", (void *)S_set_enable_object_counts);
@@ -197,6 +212,8 @@ void S_prim_init() {
     Sforeign_symbol("(cs)enable_object_backreferences", (void *)S_enable_object_backreferences);
     Sforeign_symbol("(cs)set_enable_object_backreferences", (void *)S_set_enable_object_backreferences);
     Sforeign_symbol("(cs)object_backreferences", (void *)S_object_backreferences);
+    Sforeign_symbol("(cs)list_bits_ref", (void *)S_list_bits_ref);
+    Sforeign_symbol("(cs)list_bits_set", (void *)S_list_bits_set);
 }
 
 static void s_instantiate_code_object() {
@@ -211,9 +228,11 @@ static void s_instantiate_code_object() {
     cookie = S_get_scheme_arg(tc, 2);
     proc = S_get_scheme_arg(tc, 3);
 
-    tc_mutex_acquire()
+    S_thread_start_code_write();
+
     new = S_code(tc, CODETYPE(old), CODELEN(old));
-    tc_mutex_release()
+
+    S_immobilize_object(new);
 
     oldreloc = CODERELOC(old);
     size = RELOCSIZE(oldreloc);
@@ -261,12 +280,15 @@ static void s_instantiate_code_object() {
     }
     S_flush_instruction_cache(tc);
 
+    S_thread_end_code_write();
+
     AC0(tc) = new;
 }
 
 static void s_link_code_object(co, objs) ptr co, objs; {
     ptr t; uptr a, m, n;
 
+    S_thread_start_code_write();
     t = CODERELOC(co);
     m = RELOCSIZE(t);
     a = 0;
@@ -285,6 +307,7 @@ static void s_link_code_object(co, objs) ptr co, objs; {
         S_set_code_obj("gc", RELOC_TYPE(entry), co, a, Scar(objs), item_off);
         objs = Scdr(objs);
     }
+    S_thread_end_code_write();
 }
 
 static INT s_check_heap_enabledp(void) {
