@@ -20,6 +20,21 @@
 # include <CoreFoundation/CFLocale.h>
 #endif
 
+void rktio_convert_init(rktio_t *rktio) {
+#ifdef RKTIO_USE_XLOCALE
+  rktio->locale = LC_GLOBAL_LOCALE;
+#endif
+}
+
+void rktio_convert_deinit(rktio_t *rktio) {
+#ifdef RKTIO_USE_XLOCALE
+  if (rktio->locale != LC_GLOBAL_LOCALE) {
+    freelocale(rktio->locale);
+    rktio->locale = LC_GLOBAL_LOCALE;
+  }
+#endif
+}
+
 /*============================================================*/
 /* Using iconv via a DLL                                      */
 /*============================================================*/
@@ -143,7 +158,7 @@ static void init_iconv()
       iconv_errno = (errno_proc_t)GetProcAddress(m, "_errno");
     if (!iconv_errno) {
       /* The iconv.dll distributed with Racket links to msvcrt.dll.
-	 It's a slighly dangerous assumption that whatever iconv we
+	 It's a slightly dangerous assumption that whatever iconv we
 	 found also uses msvcrt.dll. */
       m = LoadLibraryW(L"msvcrt.dll");
       if (m) {
@@ -228,6 +243,15 @@ int rktio_convert_properties(rktio_t *rktio)
 
 void rktio_set_locale(rktio_t *rktio, const char *name)
 {
+#ifdef RKTIO_USE_XLOCALE
+  if (rktio->locale != LC_GLOBAL_LOCALE) {
+    freelocale(rktio->locale);
+    rktio->locale = LC_GLOBAL_LOCALE;
+  }
+  rktio->locale = newlocale(LC_COLLATE_MASK | LC_CTYPE_MASK, name, NULL);
+  if (rktio->locale == NULL)
+    rktio->locale = LC_GLOBAL_LOCALE;
+#else
   /* We only need CTYPE and COLLATE; two calls seem to be much
      faster than one call with ALL */
   if (name) {
@@ -239,22 +263,37 @@ void rktio_set_locale(rktio_t *rktio, const char *name)
     setlocale(LC_CTYPE, "C");
     setlocale(LC_COLLATE, "C");
   }
+#endif
 }
 
-char *rktio_push_c_numeric_locale(rktio_t *rktio)
+void rktio_set_default_locale(const char *name)
 {
+  setlocale(LC_ALL, name);
+}
+
+void *rktio_push_c_numeric_locale(rktio_t *rktio)
+{
+#ifdef RKTIO_USE_XLOCALE
+  return (void *)uselocale(newlocale(LC_NUMERIC, "C", NULL));
+#else
   char *prev;
   prev = setlocale(LC_NUMERIC, NULL);
   if (!prev || !strcmp(prev, "C"))
     return NULL;
   else
     return setlocale(LC_NUMERIC, "C");
+#endif
 }
 
-void rktio_pop_c_numeric_locale(rktio_t *rktio, char *prev)
+void rktio_pop_c_numeric_locale(rktio_t *rktio, void *prev)
 {
+#ifdef RKTIO_USE_XLOCALE
+  locale_t tmp = uselocale((locale_t)prev);
+  freelocale(tmp);
+#else
   if (prev)
     setlocale(LC_NUMERIC, prev);
+#endif
 }
 
 /*============================================================*/
@@ -263,7 +302,7 @@ void rktio_pop_c_numeric_locale(rktio_t *rktio, char *prev)
 
 #ifdef RKTIO_SYSTEM_WINDOWS
 
-static char *nl_langinfo_dup()
+static char *nl_langinfo_dup(rktio_t *rktio)
 {
   int i;
   char *current_locale_name;
@@ -318,11 +357,17 @@ static char *nl_langinfo_dup()
 
 #else
 
-static char *nl_langinfo_dup()
+static char *nl_langinfo_dup(rktio_t *rktio)
 {
   char *s;
 # if HAVE_CODESET
+#  if defined(RKTIO_USE_XLOCALE)
+  locale_t old_l = uselocale(rktio->locale);
+#  endif
   s = nl_langinfo(CODESET);
+#  if defined(RKTIO_USE_XLOCALE)
+  uselocale(old_l);
+#  endif
 # else
   /* nl_langinfo doesn't work, so just make up something */
   s = "UTF-8";
@@ -335,7 +380,7 @@ static char *nl_langinfo_dup()
 
 char *rktio_locale_encoding(rktio_t *rktio)
 {
-  return nl_langinfo_dup();
+  return nl_langinfo_dup(rktio);
 }
 
 /*============================================================*/
@@ -539,7 +584,11 @@ char *rktio_locale_recase(rktio_t *rktio,
   /* Re-case chars in "out" */
   for (i = 0; i < iilen; i++) {
     char t;
+#ifdef RKTIO_USE_XLOCALE
+    t = (to_up) ? toupper_l(out[i], rktio->locale) : tolower_l(out[i], rktio->locale);
+#else
     t = (to_up) ? toupper(out[i]) : tolower(out[i]);
+#endif
     out[i] = t;
   }
 
@@ -553,6 +602,9 @@ char *rktio_locale_recase(rktio_t *rktio,
   wchar_t *wc, *ws, wcbuf[RKTIO_WC_BUF_SIZE], cwc;
   const char *s;
   unsigned int j;
+# if defined(RKTIO_USE_XLOCALE)
+  locale_t old_l = uselocale(rktio->locale);
+# endif
   /* The "n" versions are apparently not too standard: */
 # define mz_mbsnrtowcs(t, f, fl, tl, s) mbsrtowcs(t, f, tl, s)
 # define mz_wcsnrtombs(t, f, fl, tl, s) wcsrtombs(t, f, tl, s)
@@ -584,12 +636,20 @@ char *rktio_locale_recase(rktio_t *rktio,
 
   if (to_up) {
     for (j = 0; j < wl; j++) {
+# ifdef RKTIO_USE_XLOCALE
+      cwc = towupper_l(wc[j], rktio->locale);
+# else
       cwc = towupper(wc[j]);
+#endif
       wc[j] = cwc;
     }
   } else {
     for (j = 0; j < wl; j++) {
+# ifdef RKTIO_USE_XLOCALE
+      cwc = towlower_l(wc[j], rktio->locale);
+# else
       cwc = towlower(wc[j]);
+# endif
       wc[j] = cwc;
     }
   }
@@ -615,6 +675,10 @@ char *rktio_locale_recase(rktio_t *rktio,
 
   if (wc != wcbuf) free(wc);
 
+# if defined(RKTIO_USE_XLOCALE)
+  (void)uselocale(old_l);
+# endif
+
   return out;
 #endif
 }
@@ -678,7 +742,11 @@ rktio_char16_t *rktio_recase_utf16(rktio_t *rktio, rktio_bool_t to_up, rktio_cha
 
 int rktio_locale_strcoll(rktio_t *rktio, const char *s1, const char *s2)
 {
+#ifdef RKTIO_USE_XLOCALE
+  return strcoll_l(s1, s2, rktio->locale);
+#else
   return strcoll(s1, s2);
+#endif
 }
 
 int rktio_strcoll_utf16(rktio_t *rktio,
